/* region.c */

/* Copyright 2001 by Steve Kirkendall */

#include "elvis.h"

#ifdef FEATURE_REGION

# if USE_PROTOTYPES
  static region_t *rgnfind(MARK from, MARK to, _char_ font, region_t *after,
  			region_t **lag);
  static void rgnfree(region_t *region);
  static void rgnclean(BUFFER buf);
# endif


/* find an existing region containing the given text, of the given type */
static region_t *rgnfind(from, to, font, after, lag)
	MARK from;	/* start of the region to search */
	MARK to;	/* end of the region to search */
	_char_ font;	/* font to search for, or '\0' for any */
	region_t *after;/* NULL to start new search, else previous return val */
	region_t **lag;	/* reference to a pointer to region before found one */
{
	region_t *r;

	/* for each region... */
	if (lag) *lag = after;
	for (r = after ? after->next : markbuffer(from)->regions;
	     r && markoffset(r->to) <= markoffset(from);
	     r = r->next)
	{
		if (lag) *lag = r;
	}

	/* did we find a match? */
	if (r
	 && markoffset(r->from) < markoffset(to)
	 && (!font || r->font == font))
		/* Yes!  return it */
			return r;

	/* no matching region was found */
	return NULL;
}

/* free a region's resources */
static void rgnfree(region)
	region_t *region; /* the region to free */
{
	markfree(region->from);
	markfree(region->to);
	safefree(region->comment);
	safefree(region);
}


/* clean up various forms of ugliness in the region list */
static void rgnclean(buf)
	BUFFER	buf;	/* buffer whose list should be cleaned */
{
	region_t *r;	/* the new region */
	region_t *lag;	/* the region before it in the list */
	region_t *next;	/* the next region */

	/* for each region... */
	for (lag = NULL, r = buf->regions; r; lag = r, r = next)
	{
		next = r->next;

		/* if zero-length region, then delete it */
		if (markoffset(r->from) >= markoffset(r->to))
		{
			if (lag)
				lag->next = next;
			else
				buf->regions = next;
			rgnfree(r);

			/* loop, without advancing "r" */
			r = lag;
			continue;
		}

		/* if next region is same as this one, merge them */
		if (next
		 && markoffset(r->to) == markoffset(next->from)
		 && r->font == next->font
		 && !CHARcmp(r->comment, next->comment))
		{
			marksetoffset(r->to, markoffset(next->to));
			r->next = next->next;
			rgnfree(next);
			next = r->next;

			/* loop, without advancing "r" */
			r = lag;
			continue;
		}
	}
}


/* delete regions in a given range, optionally matching a given font */
void regiondel(from, to, font)
	MARK from;	/* start of region to delete */
	MARK to;	/* end of region to delete */
	_char_ font;	/* font to use, or '\0' for all fonts */
{
	region_t *doomed, *lag;

	/* for each doomed region... */
	lag = NULL;
	while ((doomed = rgnfind(from, to, font, lag, &lag)) != NULL)
	{
		/* remove it from the list */
		if (lag)
			lag->next = doomed->next;
		else
			markbuffer(from)->regions = doomed->next;

		/* if part of it extends outside the given region, add that
		 * portion back.
		 */
		if (markoffset(doomed->from) < markoffset(from))
			regionadd(doomed->from, from, doomed->font, doomed->comment);
		if (markoffset(doomed->to) > markoffset(to))
			regionadd(to, doomed->to, doomed->font, doomed->comment);

		/* free this one's resources */
		rgnfree(doomed);
	}
}

/* add a region, or extend an overlapping identical region */
void regionadd(from, to, font, comment)
	MARK from;	/* start of region to add */
	MARK to;	/* end of region to add */
	_char_ font;	/* font to use */
	CHAR *comment;	/* comment to use */
{
	region_t *r;	/* the new region */
	region_t *lag;	/* the region before it in the list */

	/* Do nothing if it would be zero-length */
	if (markoffset(from) >= markoffset(to))
		return;

	/* Delete any existing regions here.  NOTE: Although regionadd() and
	 * regiondel() are mutually recursive, there is danger of infinite
	 * recursion since regiondel() only calls regionadd() for areas ranges
	 * which are know to be free of existing regions.
	 */
	regiondel(from, to, '\0');

	/* create a new region */
	r = (region_t *)safealloc(1, sizeof(region_t));
	r->from = markdup(from);
	r->to = markdup(to);
	r->font = font;
	r->comment = CHARdup(comment);

	/* decide where to insert it */
	(void)rgnfind(from, to, '\0', NULL, &lag);

	/* insert it into the region list */
	if (lag)
	{
		r->next = lag->next;
		lag->next = r;
	}
	else
	{
		r->next = markbuffer(from)->regions;
		markbuffer(from)->regions = r;
	}

	/* clean up the list by deleting zero-length regions, and merging
	 * adjacent, identical regions.
	 */
	rgnclean(markbuffer(from));

}

/* return the region containing a given MARK, or NULL if there is none */
region_t *regionfind(mark)
	MARK	mark;	/* position, possibly within a region */
{
	MARKBUF	tmp;
	(void)marktmp(tmp, markbuffer(mark), markoffset(mark) + 1);
	return rgnfind(mark, &tmp, '\0', NULL, NULL);
}

/* This implements the :region and :unregion commands. */
RESULT ex_region(xinf)
	EXINFO	*xinf;	/* details about the command */
{
	char	font, newfont;
	CHAR	*comment;
	region_t *r;
	CHAR	buf[100], *build;
	BOOLEAN	any;

	assert(xinf->command == EX_REGION || xinf->command == EX_UNREGION);

	/* convert the font name to a font code */
	if (xinf->lhs)
		font = colorfind(xinf->lhs);
	else if (xinf->command == EX_UNREGION)
		font = '\0';
	else if (xinf->command == EX_CHREGION)
	{
		msg(MSG_ERROR, "missing lhs");
		return RESULT_ERROR;
	}
	else
	{
		/* :region with no args -- list the regions */

		/* clean it first */
		rgnclean(markbuffer(xinf->fromaddr));

		/* now list it */
		any = False;
		r = NULL;
		while ((r = rgnfind(xinf->fromaddr, xinf->toaddr, '\0', r, NULL)) != NULL)
		{
			/* list this region */
			any = True;
			long2CHAR(buf, markline(r->from));
			build = buf + CHARlen(buf);
			*build++ = ',';
			long2CHAR(build, markline(r->to) - 1);
			CHARcat(build, toCHAR(" region "));
			CHARcat(build, colorinfo[(int)r->font].name);
			build += CHARlen(build);
			if (CHARcmp(r->comment, colorinfo[(int)r->font].name))
			{
				*build++ = ' ';
				CHARcat(build,  r->comment);
				build += CHARlen(build);
			}
			*build++ = '\n';
			drawextext(xinf->window, buf, (int)(build - buf));
		}
		if (!any)
			msg(MSG_INFO, "no regions");
		return RESULT_COMPLETE;
	}

	/* default comment is same as face name */
	comment = xinf->rhs ? xinf->rhs : xinf->lhs;

	/* add or delete regions */
	if (xinf->command == EX_REGION)
		regionadd(xinf->fromaddr, xinf->toaddr, font, comment);
	else if (xinf->command == EX_UNREGION)
		regiondel(xinf->fromaddr, xinf->toaddr, font);
	else /* EX_CHREGION */
	{
		/* parse the newfont and comment */
		if (!xinf->rhs)
		{
			msg(MSG_ERROR, "missing rhs");
			return RESULT_ERROR;
		}
		newfont = colorfind(xinf->rhs);
		comment = CHARchr(comment, ' ');
		if (comment)
		{
			*comment = '\0';
			while (*comment == ' ');
				comment++;
		}
		else
			comment = xinf->rhs;

		/* for each matching region... */
		r = NULL;
		while ((r = rgnfind(xinf->fromaddr, xinf->toaddr, font, r, NULL)) != NULL)
		{
			/* change the font & comment */
			r->font = newfont;
			safefree(r->comment);
			r->comment = CHARdup(comment);
		}
	}

	return RESULT_COMPLETE;
}

#endif /* FEATURE_REGION */
